"""
FtpCube
Copyright (C) Michael Gilfix

This file is part of FtpCube.

You should have received a file COPYING containing license terms
along with this program; if not, write to Michael Gilfix
(mgilfix@eecs.tufts.edu) for a copy.

This version of FtpCube is open source; you can redistribute it and/or
modify it under the terms listed in the file COPYING.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
"""

import os

class ArchiveHandler:
    """Base class for all archive handlers.

    Archive handlers are implementations class that know how to unpack and process
    archives with certain extensions. Archive handlers take a reference to a parent
    window, in which they can create a progress dialog that indicates progress
    when processing an archive. There should be one handler instance per archive
    being processed."""

    def __init__(self, **kwargs):
        """Creates a new archive handler.

        The following keyword arguments are defined:
            'path'   : The path to the archive file
            'parent' : The parent window for the progress dialog
        """
        try:
            self.path = kwargs['path']
        except KeyError:
            self.path = None

        try:
            self.parent = kwargs['parent']
        except KeyError:
            self.parent = None

    def getPath(self):
        """Returns the path of the archive associated with this handler."""
        return self.path

    def setPath(self, path):
        """Sets the path of the archive associated with this handler."""
        self.path = path

    def getParentWindow(self):
        """Gets the parent window for the handler's progress bar."""
        return self.parent

    def setParentWindow(self, window):
        """Sets the parent window for the handler's progress bar."""
        self.parent = window

    def isValid(self):
        """Returns a boolean indicated whether the archive for this handler is valid.

        The means of determining whether an archive is valid is archive-type dependent.
        Validation may be more accurate for some archive types than others."""
        return False

    def getDescription(self):
        """Returns a description of the type of archive being handled.

        This description should be in upper case letters. For archive handlers that
        handle multiple types, this description should match the file being handled."""
        return None

    def unpack(self, location):
        """Unpacks archive contents into the specified location directory.
        
        The location should be a directory. If it does not exist, then the directory is
        automatically created. If location points to a file, then a ValueError exception
        is raised."""
        if not os.path.exists(location):
            try:
                os.mkdir(location)
                os.chmod(location, 0700)
            except OSError, strerror:
                raise
        elif not os.path.isdir(location):
            raise ValueError, location