"""
FtpCube
Copyright (C) Michael Gilfix

This file is part of FtpCube.

You should have received a file COPYING containing license terms
along with this program; if not, write to Michael Gilfix
(mgilfix@eecs.tufts.edu) for a copy.

This version of FtpCube is open source; you can redistribute it and/or
modify it under the terms listed in the file COPYING.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
"""

import libftpcube.utils
import archhandler

import wx

import os
import tarfile

class TarHandler(archhandler.ArchiveHandler):
    """Archive hadnelr for TAR files.

    This archive handler can process compressed tar files in the common compression
    formats: gzip and bzip2. The appropriate uncompression method is used automatically
    by the handler."""

    def __init__(self, **kwargs):
        """Creates a new TAR archive handler instance and associates it with the
        archive file specified as a keyword argument."""
        archhandler.ArchiveHandler.__init__(self, **kwargs)

    def isValid(self):
        """Returns true if the TAR archive is valid.

        The tar API offers limited validation capability without attempting to
        extract the files. The validation routine attempts to retrieve the list
        of file members in the archive. If this list can be retrieved, then the
        archive is considered valid."""
        valid = False
        path = self.getPath()
        if libftpcube.utils.isStrOrUnicode(path):
            tar = None
            try:
                # Better validation could go here in the future
                tar = tarfile.open(path, 'r')
                valid = True
            finally:
                try:
                    if tar:
                        tar.close()
                except Exception:
                    pass
        return valid

    def getDescription(self):
        """Returns an upper case description for the tar file.

        This description is determined dynamically based on the extension of the
        archive file."""
        desc = None
        path = self.getPath()
        path = path.lower()
        for ext in getExtensions():
            if path.endswith(ext):
                desc = ext[1:].upper()
                break
        if desc is None:
            desc = 'TAR'
        return desc

    def unpack(self, location):
        """Unpacks the associated TAR archive into the specified location.
    
        The location should be a directory. If it does not exist, then the directory is
        automatically created. If location points to a file, then a ValueError exception
        is raised."""
        archhandler.ArchiveHandler.unpack(self, location)

        tar_file = None
        progress = None
        file_list = [ ]
        try:
            try:
                path = self.getPath()
                tar_file = tarfile.open(path, 'r')
                members = tar_file.getmembers()
                progress = wx.ProgressDialog(
                    _("Ftpcube - Unarchiving"),
                    _("Untarring %(file)s") %{ 'file' : os.path.basename(path) },
                    len(members),
                    self.getParentWindow(),
                    style=wx.PD_APP_MODAL | wx.PD_ESTIMATED_TIME | wx.PD_REMAINING_TIME |
                          wx.PD_CAN_ABORT)
                for i in range(len(members)):
                    member = members[i]
                    if not member.isdir():
                        extract_path = os.path.normpath(os.path.join(location, member.name))
                        file_list.append(extract_path)
                    tar_file.extract(member, location)
                    # Update every 8th file
                    if i % 8 == 0 and not progress.Update(i):
                        break
                progress.Update(i)
            except Exception, strerror:
                raise
        finally:
            if tar_file:
                try:
                    tar_file.close()
                except Exception:
                    pass
            if progress:
                progress.Destroy()
        return file_list

def getExtensions():
    """Returns a list of extensions supported by this handler."""
    return [ '.tar', '.tar.gz', '.tar.Z', '.tar.bz2' ]

def getHandlerClass():
    """Returns the handler class instance."""
    return TarHandler
