"""
FtpCube
Copyright (C) Michael Gilfix

This file is part of FtpCube.

You should have received a file COPYING containing license terms
along with this program; if not, write to Michael Gilfix
(mgilfix@eecs.tufts.edu) for a copy.

This version of FtpCube is open source; you can redistribute it and/or
modify it under the terms listed in the file COPYING.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
"""

import libftpcube.utils
import archhandler

import wx

import os
import zipfile

class ZipHandler(archhandler.ArchiveHandler):
    """Archive handler for ZIP files."""

    def __init__(self, **kwargs):
        """Creates a new ZIP archive handler instance and associates it with the archive
        file specified as a keyword argument."""
        archhandler.ArchiveHandler.__init__(self, **kwargs)

    def isValid(self):
        """Returns true if the ZIP file associated with this handler is valid.

        A valid ZIP file consists of an archive where all file contents pass their CRC
        checks."""
        valid = False
        path = self.getPath()
        if libftpcube.utils.isStrOrUnicode(path) and zipfile.is_zipfile(path):
            zip = None
            try:
                zip = zipfile.ZipFile(path, 'r')
                if zip.testzip() is None:
                    valid = True
            finally:
                try:
                    if zip:
                        zip.close()
                except Exception:
                    pass
        return valid

    def getDescription(self):
        """Returns the string 'ZIP'"""
        return "ZIP"

    def unpack(self, location):
        """Unpacks the associated ZIP archive into the specified location.

        The location should be a directory. If it does not exist, then the directory is
        automatically created. If location points to a file, then a ValueError exception
        is raised."""
        archhandler.ArchiveHandler.unpack(self, location)

        zip_file = None
        progress = None
        file_list = [ ]
        try:
            try:
                path = self.getPath()
                zip_file = zipfile.ZipFile(path, 'r')
                names = zip_file.namelist()
                progress = wx.ProgressDialog(
                    _("Ftpcube - Unarchiving"),
                    _("Unzipping %(file)s") %{ 'file' : os.path.basename(path) },
                    len(names),
                    self.getParentWindow(),
                    style=wx.PD_APP_MODAL | wx.PD_ESTIMATED_TIME | wx.PD_REMAINING_TIME |
                          wx.PD_CAN_ABORT)
                for i in range(len(names)):
                    name = names[i]
                    info = zip_file.getinfo(name)
                    writeable = os.path.normpath(os.path.join(location, name))

                    base_dir = os.path.dirname(writeable)
                    if not os.path.exists(base_dir):
                        os.mkdir(base_dir)
                        os.chmod(base_dir, 0700)

                    if writeable[-1] in '/\\':
                        os.mkdir(writeable)
                        os.chmod(writeable, 0700)                        
                    else:
                        file = open(writeable, 'w')
                        try:
                            bytes = zip_file.read(name)
                            file.write(bytes)
                        finally:
                            file.close()
                        file_list.append(writeable)
                    # Update every 8th file
                    if i % 8 == 0 and not progress.Update(i):
                        break
                progress.Update(i)
            except Exception, strerror:
                raise
        finally:
            if zip_file:
                try:
                    zip_file.close()
                except Exception:
                    pass
            if progress:
                progress.Destroy()
        return file_list


def getExtensions():
    """Returns a list of extensions supported by this handler."""
    return [ '.zip' ]

def getHandlerClass():
    """Returns the handler class instance."""
    return ZipHandler